/*********************************************************************
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2022, KR Soft s.r.o.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the KR2013ApS nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************/

#ifndef ROBOT_CONTROL_ACTION_H
#define ROBOT_CONTROL_ACTION_H

#include "kr2/kord/protocol/KORDItemIDs.h"
#include "kr2/kord/protocol/JointFirmwareCommand.h"
#include "kr2/kord/protocol/DataDescriptions/Requests/ControlCommandItems.h"
#include <kr2_rc_api/api_v1/arm/trajectory.h>

#include <array>
#include <chrono>

#include <netinet/in.h>

namespace kr2 {
namespace kord {

typedef std::chrono::time_point<std::chrono::high_resolution_clock> timepoint;

template<class A, size_t B>
struct MoveBase
{
    using TTP =  kr2rc_api::Trajectory::CmdTrajectoryParams;

    public:
    void setTargets(const std::array<A, B>& a_targets)              { this->targets_        = a_targets;        }
    void setTargetsD(const std::array<A, B>& a_targets_d)           { this->targets_d_      = a_targets_d;      }
    void setTargetsDD(const std::array<A, B>& a_targets_dd)         { this->targets_dd_     = a_targets_dd;     }
    void setTargetsTorque(const std::array<A, B>& a_targets_torque) { this->targets_torque_ = a_targets_torque; }

    std::array<A, B> getTargets()       const { return targets_;        }
    std::array<A, B> getTargetsD()      const { return targets_d_;      }
    std::array<A, B> getTargetsDD()     const { return targets_dd_;     }
    std::array<A, B> getTargetsTorque() const { return targets_torque_; }

    void setTrackingType(const TTP::TrackingType& a_tt) { this->tt_       = a_tt;       }
    void setTrackingTypeValue(const double& a_tt_value) { this->tt_value_ = a_tt_value; }
    void setBlendType(const TTP::BlendType& a_bt)       {this->bt_        = a_bt;       }
    void setBlendTypeValue(const double& a_bt_value)    { this->bt_value_ = a_bt_value; }
    void setOverlayType(const TTP::OverlayType& a_ot)   { this->ot_       = a_ot;       }
    void setSyncTime(const double& a_sync_)             { this->sync_     = a_sync_;    }

    // set for velocity
    void setSyncValue(const long& a_sync_val)           { this->sync_val_ = a_sync_val; }
    void setPeriod(const double& a_period)              { this->period_ = a_period;     }
    void setTimeout(const double& a_timeout)              { this->timeout_ = a_timeout;     }
    //

    TTP::TrackingType getTrackingType() const { return tt_;       }
    double getTrackingTypeValue()       const { return tt_value_; }
    TTP::BlendType getBlendType()       const { return bt_;       }
    double getBlendTypeValue()          const { return bt_value_; }
    TTP::OverlayType getOverlayType()   const { return ot_;       }
    double getSyncTime()                const { return sync_;     }

    // get for velocity
    long getSyncValue()                const { return sync_val_; }
    double getPeriod()                  const { return period_;   }
    double getTimeout()                 const { return timeout_;  }
    //



    private:

        std::array<A, B> targets_;
        std::array<A, B> targets_d_;
        std::array<A, B> targets_dd_;
        std::array<A, B> targets_torque_;

        double tt_value_;
        TTP::TrackingType tt_;

        TTP::BlendType bt_;
        double bt_value_;

        TTP::OverlayType ot_;

        double sync_;

        // for velocity
        long sync_val_;
        double period_;
        double timeout_;
        //

};

/// @brief Holds references that are used for controlling the robot movement.
///        Robot movement can be controlled by providing the references for
///        joint space, work space, and brakes.
///
///        It is obtained from the command frame.
class RobotControl {
    public:
        RobotControl(){
#ifdef __APPLE__
            // TODO: the cast should be subjected for some investigation
            last_fw_cmd_processed_ = std::chrono::steady_clock::now();
#else
            last_fw_cmd_processed_ = std::chrono::system_clock::now();
#endif
        }

        kr2::kord::protocol::EKORDItemID item_id_;
        timepoint capture_ts_;

        uint16_t sequence_num_ = 0;

        MoveBase<double, 6UL> move_l_;
        MoveBase<double, 6UL> move_l_w_dyn_;
        MoveBase<double, 7UL> move_j_;

        // Remote host ip
        sockaddr_in remote_host_;

        uint8_t v_    = 0;
        int64_t mask_ = 0;
        
        timepoint last_fw_cmd_processed_;
        std::vector<std::pair<unsigned int, kr2::kord::protocol::EJointFirmwareCommand>> jnt_fw_commands_;
        int64_t jnt_fw_cmd_frame_stamp_ = 0;

        uint64_t time_stamp_token_;

        uint64_t composed_dependent_; // to keep of one command composed of multiple others interpreted as one

        uint64_t failed_by_not_completed = 1; // 1 is nothing

        bool crc_error_;
        // for Frame
        std::array<double, 6UL> pose_;
        unsigned int frame_id_;
        unsigned int ref_frame_;
        // for Load
        double load_mass_;
        unsigned int load_id_;
        std::array<double, 3UL> load_cog_;
        std::array<double, 6UL> load_inertia_;
        // for Alarm
        unsigned int alarm_id_;

        //
        // RequestSystem content Items
        //
        // If the RequestSystem ContentItemID is captured, the undelying command is stored here
        kr2::kord::protocol::EControlCommandItems req_system_id_ = kr2::kord::protocol::EControlCommandItems::eInvalid;

        //
        // RequestTransfer content Items
        //
        unsigned int transfer_mask_;

        // Unique identifier of the command - 64bit timestamp
        int64_t req_sys_tx = 0;

        // Status Echo

        // Echoed CBun's status timestamp by API
        int64_t status_echo_tx = 0;

        // API's sent timestamp
        int64_t status_echo_rx = 0;

};

} // kord
} // kr2

#endif // ROBOT_CONTROL_ACTION_H
