
#ifndef KR2KORD_EXTRAPOLATOR_H
#define KR2KORD_EXTRAPOLATOR_H

#include <vector>
#include <algorithm>
#include <deque>

namespace kr2 {
namespace utils {

//No noise is assumed, no filtering done.
class DSAExtrapolator {
    public:
        DSAExtrapolator():
            sampling_time_(0.0),
            distance_(0.0),
            speed_(0.0),
            acceleration_(0.0)
        {}

        DSAExtrapolator(double sampling_time_seconds):
            sampling_time_(sampling_time_seconds),
            distance_(0.0),
            speed_(0.0),
            acceleration_(0.0)
        {}

        //! dist[m], vel[m/s], acc[m/s^2]
        void update(double dist, double spd, double acc) {
            distance_ = dist;
            speed_ = spd;
            acceleration_ = acc;
        }

        //! Steps are number of samples into the future
        double extrapolate(int steps) const {
            double t = sampling_time_ * steps;

            double edst = distance_;
            edst += speed_ * t;
            edst += 0.5 * acceleration_ * t * t;

            return edst;
        }

        void reset() {
            distance_ = 0.0;
            speed_ = 0.0;
            acceleration_ = 0.0;
        }

    protected:
        double sampling_time_;
        
        double distance_;
        double speed_;
        double acceleration_; 

};

class DExtrapolator : public DSAExtrapolator {
public:
    DExtrapolator():
        DSAExtrapolator(0),
        window_size_(0)
    {}

    DExtrapolator(double sampling_time):
        DSAExtrapolator(sampling_time),
        window_size_(3)
    {}

    void updateDistance(double d) {
        window_.push_back(d);
        if (window_.size() > window_size_){
            window_.pop_front();
        }

        if (window_.empty()) { return; }

        // update distane
        if (window_.size() >= 1) { 
            this->distance_ = window_.back();
        }

        // estimate speed
        if (window_.size() >= 2) {
            this->speed_ = (*(window_.end()-1) - *(window_.end()-2));
            this->speed_ /= this->sampling_time_;
        }

        // estimate acceleration
        if (window_.size() >= 3){
            this->acceleration_ = (window_.at(2) - window_.at(1)) -(window_.at(1) - window_.at(0));
            this->acceleration_ /= this->sampling_time_;
        }
    }

    void reset() {
        window_.clear();
        DSAExtrapolator::reset();
    }

private:
    std::deque<double> window_;
    int window_size_;
};

class VectorExtrapolator {
public:
    VectorExtrapolator(double sampling_time_seconds, size_t width):
        counter_(0)
    {
        DExtrapolator se(sampling_time_seconds);
        extrapolation_.resize(width);
        std::fill(extrapolation_.begin(), extrapolation_.end(), se);
    }

    //! Update the values to start the extrapolation from.
    //! Automatically reset if the extrapolation was previously active
    void update(const std::vector<double> &d, const std::vector<double> &s, const std::vector<double> &a){
        int i = 0;
        if (counter_){
            reset();
        }
        for (auto &e : extrapolation_){
            e.update(d.at(i), s.at(i), a.at(i)); i++;
        }
    }

    //! Automatically reset if the extrapolation was previously active
    void update(const std::vector<double> &d){
        int i = 0;
        if (counter_){
            reset();
        }
        for (auto &e : extrapolation_){
            e.updateDistance(d.at(i)); i++;
        }
    }

    //! Provide extrapolation and if no extrapolation is possible, then return latest updated value.
    //! Count the amount of steps to extrapolate internally.
    std::vector<double> extrapolate(){
        std::vector<double> vec;

        counter_++;
        if (counter_ > MAX_STEPS_TO_EXTRAPOLATE_) {
            // Do it i a better way: stop the robot safely with splines
            counter_ = MAX_STEPS_TO_EXTRAPOLATE_;  // It may currently produce EStop/TRQ dev
        }
        vec = extrapolate(counter_);

        return vec;
    }

    //! Extrapolate given number of steps to the future.
    std::vector<double> extrapolate(int i) const {
        std::vector<double> extrapolated_;

        for (const auto &it : extrapolation_){
            extrapolated_.push_back(it.extrapolate(i));
        }

        return extrapolated_;
    }

    void reset() {
        counter_ = 0;
        for (auto &e : extrapolation_){
            e.reset();
        }
    }

    int counter() const {
        return counter_;
    }

private:
    std::vector<DExtrapolator> extrapolation_;
    int counter_;
    static const int MAX_STEPS_TO_EXTRAPOLATE_ = 5;
};

} // kswx_kr2_kord
} // utils

#endif // KR2KORD_EXTRAPOLATOR_H
